#!/usr/bin/env python

"""
<par>This module implements <app>make</app> functionality.</par>

<par>Like <app>make</app> it allows you to specify dependencies
between files and actions to be executed when files don't exist or are out of date
with respect to one of their sources. But unlike <app>make</app> you
can do this in a object oriented way and targets are not only limited to files,
but you can implement e.g. dependencies on database records by implementing new
<pyref class="Target"><class>Target</class></pyref> classes.</par>

<par>Relevant classes are:
<ulist>
<item><pyref class="Project"><class>Project</class></pyref>, which is the container for all targets,
actions and dependencies in a project,</item>
<item><pyref class="Target"><class>Target</class></pyref> (and subclasses), which are files
(or other entities like database records) that have to be built by actions,</item>
<item><pyref class="Dep"><class>Dep</class></pyref> (and subclasses) which specify types
of dependencies between targets and</item>
<item><pyref class="Action"><class>Action</class></pyref> (and subclasses) which specify how targets
are (re)built if they don't exist or aren't up to date.</item>
</ulist>
</par>
"""

import sys, os, os.path, time, getopt, types, warnings, re

from mx import DateTime

from ll import ansistyle, url

indent = " "*8

class MakeWarning(Warning):
	pass

class RedefinedTargetWarning(MakeWarning):
	def __init__(self, targetid):
		self.targetid = targetid
	def __str__(self):
		return "target with id=%r redefined" % self.targetid

class ClockSkewWarning(MakeWarning):
	def __init__(self, targetid, lastmodified, now):
		self.targetid = targetid
		self.lastmodified = lastmodified
		self.now = now
	def __str__(self):
		return "target %r is from the future: %r>%r=now" % (self.targetid, self.lastmodified, self.now)

class Dep(object):
	"""
	<par>A <class>Dep</class> is a dependency between a <z>source</z>
	and a <z>target</z> <pyref class="Target"><class>Target</class></pyref>. It's possible
	to assign specific meanings to certain dependencies by using subclasses of
	<class>Dep</class>.</par>
	"""
	__slots__ = ("source", "target")

	def __init__(self, source, target):
		"""
		<par>create a <class>Dep</class> instance.</par>
		"""
		self.source = source
		self.target = target
		target._sources.append(self)
		source._targets.append(self)

	def __hash__(self):
		return hash(self.source.id) ^ hash(self.target.id)

	def __cmp__(self, other):
		return cmp(self.source.id, other.source.id) or cmp(self.target.id, other.target.id)

	def __repr__(self):
		return "<%s from %r to %r at 0x%x>" % (self.__class__.__name__, self.source, self.target, id(self))

class MainDep(Dep):
	"""
	<par>This class can be used to create a special dependency between
	too targets. This is e.g. used by <pyref class="CopyAction"><class>CopyAction</class></pyref>
	and <pyref class="XISTAction"><class>XISTAction</class></pyref>.</par>
	"""
	pass

class OrderedDep(MainDep):
	"""
	<par>This dependency class represents an ordered dependency, i.e. each
	instance has a order attribute.</par>
	"""
	order = 0

	__slots__ = ("order")

	def __init__(self, source, target, order=None):
		"""
		<par>create a new <class>OrderedDep</class> instance.
		When <arg>order</arg> is not specified a new unique order value
		will be used that will be greater than the order value of
		previously created <class>OrderedDep</class>
		instances.</par>
		"""
		MainDep.__init__(self, source, target)
		if order is None:
			order = self.__class__.order
			self.__class__.order += 1
		self.order = order

class Target(object):
	def __init__(self, project, id, action=None):
		"""
		<par>Create a new <class>Target</class> instance.</par>
		"""
		self.project = project
		self.id = id
		self.action = action
		self._sources = []
		self._targets = []
		self.project[self.id] = self

	def lastModified(self):
		"""
		<par>return the timestamp, when this target was last modified.
		The timestamp will be cached, as <function>os.path.getmtime</function> is quite
		slow on Windows and will be updated when a target is remade.</par>
		"""
		try:
			return self._lastModified
		except AttributeError:
			lm = self._fetchLastModified()
			now = DateTime.now()
			if lm>now:
				self.project.reportWarning(ClockSkewWarning(self, lm, now), 2)
			self._lastModified = lm
			return lm

	def _fetchLastModified(self):
		"""
		<par>return the timestamp of the last modification.
		Overwrite in subclasses.</par>
		"""
		raise NotImplementedError()

	def touch(self):
		"""
		<par>set the <z>last modified</z> timestamp to now.</par>
		"""
		self._lastModified = DateTime.now()

	def update(self):
		"""
		<par>Update this target and all of its sources, if necessary.</par>
		"""
		t1 = DateTime.now()
		for dep in self._sources:
			dep.source.update()
		if self.hasToBeRebuilt():
			if self.action is None:
				raise ValueError("don't know how to build %s" % self)
			if self.project.showTargets:
				msg = str(self)
				if self.project.showActions:
					msg += " <-%s--" % self.project.actionStr(self.action)
				elif self.project.showSources:
					msg += " <--"
				if self.project.showSources:
					if self.lastModified() is None:
						reason = "(target doesn't exist)"
					else:
						newer = [ str(dep.source) for dep in self.newerSources() ]
						if len(newer):
							if len(newer)>self.project.maxNewerReport:
								reason = "(%d sources changed)" % len(newer)
							else:
								reason = ", ".join(newer)
						else:
							reason = ""
					msg += " %s" % reason
				msg += " ...\n"
				self.project.report(msg)
			failed = 0
			try:
				msg = self.action.execute(self)
			except (SystemExit, KeyboardInterrupt):
				raise
			except Exception, ex:
				if self.project.ignoreExceptions:
					failed = 1
				else:
					raise
			t2 = DateTime.now()
			if failed:
				import traceback
				traceback.print_exc()
				fullmsg = "%s%s after %s\n" % (indent, self.project.errorStr("failed"), self.project.secondsStr(t2-t1))
				self.project.reportError(fullmsg)
				self.project.failed += 1
			else:
				if self.project.showMessages:
					fullmsg = "%sdone after %s" % (indent, self.project.secondsStr(t2-t1))
					if msg is not None:
						fullmsg += ": %s" % msg
					fullmsg += "\n"
					self.project.report(fullmsg)
				self.project.rebuilt += 1
			self.touch() # always touch the target, so building can continue even if this target failed

	def sources(self, type=Dep):
		"""
		<par>return a list of all the source dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return [ dep for dep in self._sources if isinstance(dep, type) ]

	def targets(self, type=Dep):
		"""
		<par>return a list of all the target dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return [ dep for dep in self._targets if isinstance(dep, type) ]

	def newerSources(self, type=Dep):
		"""
		<par>return a list of all the sources that or newer the this one
		(or don't exist).</par>
		"""
		selfLM = self.lastModified()
		sources = self.sources(type)
		if selfLM is None:
			return sources
		else:
			newer = []
			for dep in sources:
				sourceLM = dep.source.lastModified()
				if not sourceLM or sourceLM>selfLM:
					newer.append(dep)
			return newer

	def dependOn(self, *others):
		"""
		<par>add one or more dependencies.</par>

		<par><arg>others</arg> may contain <pyref class="Target"><class>Target</class></pyref>
		instances and subclasses of <pyref class="Dep"><class>Dep</class></pyref>. This will be used to
		create specific dependencies. Initially <pyref class="Dep"><class>Dep</class></pyref> will be used to
		construct source dependencies for every <pyref class="Target"><class>Target</class></pyref> instance
		encountered in the argument list. From the point in the argument list where a
		<pyref class="Dep"><class>Dep</class></pyref> subclass is encountered this new class will be
		used for constructing dependencies. This makes it possible to switch between different
		dependency classes.</par>
		"""
		dependencyClass = Dep
		for other in others:
			if isinstance(other, type) and issubclass(other, Dep):
				dependencyClass = other
			else:
				dependency = dependencyClass(other, self)

	def hasToBeRebuilt(self):
		"""
		<par>does <self/> have to be rebuilt (either because this target does not exist, or
		it is older than one of its sources)?</par>
		"""
		selfLM = self.lastModified()
		if selfLM:
			return len(self.newerSources())>0
		else:
			return 1

	def __str__(self):
		"""
		<par>Return a nice string representation of <self/>. This is used
		in the progress report.</par>
		"""
		try:
			name = self.__class__.name
		except AttributeError:
			name = self.__class__.__name__
		return "%s(%s)" % (name, self.project.idStr(self.id))

	def __repr__(self):
		return "<%s instance id=%r at 0x%x>" % (self.__class__.__name__, self.id, id(self))

class FileSystemTarget(Target):
	"""
	<par>A target in the filesystem, i.e. a file or directory.</par>
	"""
	pass

class FileTarget(FileSystemTarget):
	"""
	<par>a file</par>
	"""
	name = "File"

	def _fetchLastModified(self):
		try:
			return self.id.openread().lastmodified
		except OSError:
			return None

class DirectoryTarget(FileSystemTarget):
	"""
	<par>a directory.</par>
	"""
	name = "Dir"

	def _fetchLastModified(self):
		# only rebuild if it doesn't exist, if it exists do nothing
		try:
			timestamp = self.id.getmtime()
		except OSError:
			return None
		return DateTime.DateTime(1900, 1, 1)

	def hasToBeRebuilt(self):
		return self.lastModified() is None

class ImageTarget(FileTarget):
	"""
	<par>an image file.</par>
	"""
	name = "Image"

class JavascriptTarget(FileTarget):
	"""
	<par>a Javascript source file (extension <lit>.js</lit>)</par>
	"""
	name = "JS"

class CascadingStyleSheetTarget(FileTarget):
	"""
	<par>a Cascading Stylesheet file (extension <lit>.css</lit>)</par>
	"""
	name = "CSS"

class PHPTarget(FileTarget):
	"""
	<par>a PHP include file (extension <lit>.php</lit>)</par>
	"""
	name = "PHP"

class HTMLTarget(FileTarget):
	"""
	<par>a &html; file (extension <lit>.html</lit>, <lit>.phtml</lit>, etc.)</par>
	"""
	name = "HTML"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

class XMLTarget(FileTarget):
	name = "XML"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

class DTDTarget(FileTarget):
	"""
	<par>a Document Type Definition file (extension <lit>.dtd</lit>)</par>
	"""
	name = "DTD"

class PDFTarget(FileTarget):
	"""
	<par>a &pdf; file (extension <lit>.pdf</lit>)</par>
	"""
	name = "PDF"

class JavaPropTarget(FileTarget):
	"""
	<par>a Java properties file (extension <lit>.properties</lit>)</par>
	"""
	name = "JavaProp"

class PythonTarget(FileTarget):
	"""
	<par>a Python source file (extension <lit>.py</lit>)</par>
	"""
	name = "Python"

	def __init__(self, project, id, action=None, fromlist=None):
		FileTarget.__init__(self, project, id, action)
		self.fromlist = fromlist

	def load(self):
		"""
		<par>import and return the Python module as the result. (Note that
		as this method temporarily modifies <lit>sys.path</lit>, so it is not
		thread save.)</par>
		"""
		(dir, name) = os.path.split(str(self.id.asFilename()))
		(base, ext) = os.path.splitext(name)
		if base=="__init__":
			(dir, name) = os.path.split(dir)
			(base, ext) = os.path.splitext(name)
		sys.path.insert(0, dir)
		if self.fromlist is not None:
			base = ".".join(self.fromlist) + "." + base
		module = __import__(base, globals(), locals(), self.fromlist)
		del sys.path[0]
		return module

class XISTNamespaceTarget(PythonTarget):
	"""
	<par>a Python source file that contains a &xist; <pyref module="xist.xsc" class="Namespace"><class>Namespace</class></pyref>
	object and must be imported to provided element definitions etc. to the &xist; parser.</par>
	"""
	name = "XISTNS"

class XISTTarget(FileTarget):
	"""
	<par>an &xist; source file, i.e. a &xml; file that has to be transformed via &xist;
	(extensions <lit>.xsc</lit>, <lit>.pxsc</lit>, etc.)</par>
	"""
	name = "XIST"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

	def parse(self):
		from ll.xist import parsers
		return parsers.parseURL(self.id, base=self.base)

class PhonyTarget(Target):
	"""
	<par>A <class>PhonyTarget</class> does not really exist. It can
	be used as a virtual target for all the real targets that have to be built.</par>
	"""
	name = "Phony"

	def __init__(self, project, id, action=None):
		"""
		<par>create a new <class>PhonyTarget</class>.
		If <arg>action</arg> is not given or <lit>None</lit> the
		action for rebuilding <self/> will be a <pyref class="NoAction"><class>NoAction</class></pyref>
		instance.</par>
		"""
		if action is None:
			action = NoAction()
		Target.__init__(self, project, id, action)

	def _fetchLastModified(self):
		"""
		<par>get the <z>last modification</z> timestamp. This will always
		be <lit>None</lit>, as a <class>PhonyTarget</class>
		does not really exist. Note that this means that the associated action will be executed
		every time the target is updated.</par>
		"""
		return None

class Action:
	"""
	<par>An <class>Action</class> is responsible for rebuilding a <pyref class="Target"><class>Target</class></pyref>.</par>

	<par>An <class>Action</class> is stateless and so may be used by multiple targets. Executing the
	action is done through a call to the method <pyref method="execute"><method>execute</method></pyref>.</par>
	"""
	def __init__(self):
		"""
		<par>create a new <class>Action</class> instance.</par>
		"""
		pass

	def execute(self, target):
		"""
		<par>execute the action to rebuild <arg>target</arg> and
		return a one-line status/result message.</par>
		"""
		raise NotImplementedError()

	def __str__(self):
		"""
		<par>Return a string description for <self/>. This is the class name unless
		there is a class attribute <lit>name</lit>. This string is used in the progress report.</par>
		"""
		try:
			return self.__class__.name
		except AttributeError:
			return self.__class__.__name__

class CopyAction(Action):
	"""
	<par>generate the target by copying the source.</par>
	<par>The <z>source</z> is the only <pyref class="Target"><class>Target</class></pyref> on which
	the target depends through a <pyref class="MainDep"><class>MainDep</class></pyref>.</par>
	"""
	name = "Copy"

	def execute(self, target):
		"""
		<par>execute the action</par>
		"""
		deps = target.sources(MainDep)
		assert len(deps)==1, "copy must have one main source for target %r, has %r" % (target, deps)

		infile = deps[0].source.id.openread()
		data = infile.read()
		infile.close()
		outfile = target.id.openwrite()
		outfile.write(data)
		outfile.close()
		return "%s copied" % target.project.sizeStr(len(data))

class JoinAction(Action):
	"""
	<par>generate the target by joining the sources.</par>
	<par>The <z>sources</z> must depend on the target via a
	<pyref class="OrderedDep"><class>OrderedDep</class></pyref> and will be join
	in the order specified by these dependencies.</par>
	"""
	name = "Join"

	def execute(self, target):
		"""
		<par>execute the action</par>
		"""
		deps = target.sources(OrderedDep)
		deps.sort(lambda s1, s2: cmp(s1.order, s2.order))

		outfile = target.id.openwrite()
		size = 0
		for dep in deps:
			infile = dep.source.id.openread()
			data = infile.read()
			infile.close()
			outfile.write(data)
			size += len(data)
		outfile.close()
		return "%s joined" % target.project.sizeStr(size)

class MkDirAction(Action):
	"""
	<par>This action creates the target as a directory.</par>
	"""
	name = "MkDir"

	def execute(self, target):
		"""
		<par>create the directory.</par>
		"""
		os.makedirs(target.id.asFilename())

class XISTAction(Action):
	"""
	<par>This action creates the target via transforming the main source
	via <pyref module="xist"><app>&xist;</app></pyref>.</par>
	"""
	name = "XIST"

	def __init__(self, encoding=None, xhtml=None, mode=None, target=None, stage=None, lang=None, targetroot=None):
		"""
		<par>create a new <class>XISTAction</class> instance. The arguments
		will be used during parsing, transformation and publishing in the usual way.</par>
		"""
		self.encoding = encoding
		self.xhtml = xhtml
		self.mode = mode
		self.target = target
		self.stage = stage
		self.lang = lang
		self.targetroot = targetroot

	def loadNamespaces(self, target):
		"""
		<par>import all the source namespaces (i.e. all the sources of <self/> that are instances
		of <pyref class="XISTNamespaceTarget"><class>XISTNamespaceTarget</class></pyref>).</par>
		<par>The module objects will be returned in a list.</par>
		"""
		namespaces = [ dep.source.load() for dep in target.sources() if isinstance(dep.source, XISTNamespaceTarget) ]
		return namespaces

	def converter(self, maketarget=None):
		"""
		<par>create a new <pyref module="xist.converters" class="Converter"><class>Converter</class></pyref> instance
		for this action. The attributes of this new converter (<lit>mode</lit>, <lit>target</lit>,
		<lit>stage</lit>, etc.) will correspond to those specified in the constructor.</par>
		<par>When the argument <arg>target</arg> is specified it will be used to initialize the
		attribute <lit>maketarget</lit> of the converter.</par>
		"""
		from ll.xist import converters
		return converters.Converter(root=self.targetroot, mode=self.mode, stage=self.stage, target=self.target, lang=self.lang, makeaction=self, maketarget=maketarget)

	def publish(self, target, node):
		"""
		<par>publish the <arg>node</arg> as the content of the
		target <arg>target</arg>.</par>
		"""
		from ll.xist import publishers
		outfile = target.id.openwrite()
		publisher = publishers.FilePublisher(outfile, base=target.base, root=self.targetroot, encoding=self.encoding, xhtml=self.xhtml)
		publisher.doPublication(node)
		bytes = publisher.tell()
		outfile.close()
		return bytes

	def execute(self, target):
		"""
		<par>execute the action. The <arg>target</arg> must have exactly one source
		with a <pyref class="MainDep"><class>MainDep</class></pyref> dependency. This target will be transformed
		using all the namespaces available via <pyref method="loadNamespaces"><method>loadNamespaces</method></pyref>
		and the resulting node will be published via <pyref method="publish"><method>publish</method></pyref>.</par>
		"""
		self.loadNamespaces(target)
		xistdeps = target.sources(MainDep)
		assert len(xistdeps)==1, "XISTAction must have one main source for target %r, has %r" % (target, xistdeps)
		t1 = DateTime.now()
		node = xistdeps[0].source.parse()
		t2 = DateTime.now()
		converter = self.converter(maketarget=target)
		node = node.convert(converter)
		t3 = DateTime.now()
		bytes = self.publish(target, node)
		t4 = DateTime.now()
		return "parse %s; convert %s; save %s; %s published" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

	def __str__(self):
		options = []
		for optionname in ["encoding", "xhtml", "mode", "target", "stage", "lang", "targetroot"]:
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append("%s=%r" % (optionname, optionvalue))
		return "%s(%s)" % (Action.__str__(self), ", ".join(options))

class SplatAction(Action):
	"""
	<par>This action creates the target via replacing certain regular expressions
	in the main source.</par>
	"""
	name = "Splat"

	def __init__(self, patterns):
		"""
		<par>create a new <class>SplatAction</class> instance. <arg>patterns</arg>
		is a list of pattern pairs. The first entry will be replaced by the second entry.</par>
		"""
		self.patterns = patterns

	def execute(self, target):
		"""
		<par>execute the action. The <arg>target</arg> must have exactly one source
		with a <pyref class="MainDep"><class>MainDep</class></pyref> dependency. This target will be transformed
		using all the namespaces available via <pyref method="loadNamespaces"><method>loadNamespaces</method></pyref>
		and the resulting node will be published via <pyref method="publish"><method>publish</method></pyref>.</par>
		"""
		deps = target.sources(MainDep)
		assert len(deps)==1, "SplatAction must have one main source for target %r, has %r" % (target, deps)
		t1 = DateTime.now()
		data = deps[0].source.id.openread().read()
		t2 = DateTime.now()
		for (search, replace) in self.patterns:
			data = re.sub(search, replace, data)
		t3 = DateTime.now()
		file = target.id.openwrite()
		file.write(data)
		bytes = file.tell()
		file.close()
		t4 = DateTime.now()
		return "read %s; replace %s; save %s; %s written" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

class NoAction(Action):
	"""
	<par>This action does nothing.</par>
	"""
	name = "None"

	def execute(self, target):
		"""
		<par>do nothing</par>
		"""
		pass

class CommandAction(Action):
	"""
	<par>This action executes a system command (via <function>os.system</function>).</par>
	"""
	name = "Command"

	def __init__(self, command):
		"""
		<par>create a new <class>CommandAction</class>. <arg>command</arg> will be the command
		that is executed when calling <pyref method="execute"><method>execute</method></pyref>.</par>
		"""
		Action.__init__(self)
		self.command = command

	def execute(self, target):
		"""
		<par>execute the action (note that the <arg>target</arg> is irrelavant and will be ignored;
		in most cases it will probably be a <pyref class="PhonyTarget"><class>PhonyTarget</class></pyref>).</par>
		"""
		os.system(self.command)

	def __str__(self):
		return "%s(%r)" % (Action.__str__(self), self.command)

class DBID(object):
	"""
	<par>This class provides a unique identifier for database content. This
	can be used as an id for <pyref class="Target"><class>Target</class></pyref>
	objects that are not files, but database records.</par>
	"""
	def __init__(self, type, host=None, port=None, user=None, database=None, table=None, key=None):
		self.type = type
		self.host = host
		self.port = port
		self.user = user
		self.database = database
		self.table = table
		self.key = key

	def __eq__(self, other):
		res = self.__class__ == other.__class__
		if not res:
			res = self.type==other.type and self.host==other.host and self.port==other.port and self.user==other.user and self.database==other.database and self.table==other.table and self.key==other.key
		return res

	def __hash__(self):
		return hash(self.type) ^ hash(self.host) ^ hash(self.port) ^ hash(self.user) ^ hash(self.database) ^ hash(self.table) ^ hash(self.key)

	def __repr__(self):
		args = []
		for attrname in ("type", "host", "port", "user", "database", "table", "key"):
			attrvalue = getattr(self, attrname)
			if attrvalue is not None:
				args.append("%s=%r" % (attrname, attrvalue))
		return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

class Project(object):
	"""
	<par>A <class>Project</class> collects all <pyref class="Target"><class>Target</class></pyref>s
	from a project. It is responsible for initiating the build process and for generating a report about
	the progress of the build process.</par>
	"""
	def __init__(self, showSummary=1, showTargets=1, showActions=0, showSources=0, showMessages=0, ignoreExceptions=0, color=None, maxNewerReport=2):
		self.targets = {}
		self.rebuilt = 0
		self.failed = 0
		self.showSummary = showSummary
		self.showTargets = showTargets
		self.showActions = showActions
		self.showSources = showSources
		self.showMessages = showMessages
		self.ignoreExceptions = ignoreExceptions
		if color is None:
			color = os.getenv("MAKE_REPRANSI")
			if color is None:
				color = 0
			else:
				color = int(color)
				if not sys.platform.startswith("linux") and not sys.platform.startswith("darwin"):
					color = 0
		self.color = color
		self.maxNewerReport = maxNewerReport
		self._initColors()

	def _initColors(self):
		"""
		<par>initialize the color classes which are used for the various formatting
		methods.</par>
		"""
		self.TextTime = ansistyle.Text5
		self.TextSize = ansistyle.Text4
		self.TextID = ansistyle.Text2
		self.TextAction = ansistyle.Text15
		self.TextError = ansistyle.Text9

	def secondsStr(self, delta):
		"""
		<par>return a nicely formatted and colored string for
		the <class>DateTime</class> value <arg>delta</arg>.</par>
		"""
		if self.color:
			return ansistyle.Text(self.TextTime("%.03f" % delta.seconds), "s")
		else:
			return "%.03fs" % delta.seconds

	def sizeStr(self, bytes):
		"""
		<par>return a nicely formatted and colored string for
		the byte count value <arg>bytes</arg>.</par>
		"""
		if self.color:
			return ansistyle.Text(self.TextSize("%d" % bytes), " bytes")
		else:
			return "%d bytes" % bytes

	def actionStr(self, action):
		"""
		<par>return a nicely formatted and colored string for
		the action <arg>action</arg>.</par>
		"""
		s = str(action)
		if self.color:
			s = self.TextAction(s)
		return s

	def errorStr(self, text):
		"""
		<par>return a nicely formatted and colored string for
		the error text <arg>text</arg>.</par>
		"""
		if self.color:
			return self.TextError(text)
		else:
			return text

	def idStr(self, id):
		"""
		<par>return a nicely formatted and colored string for
		the id <arg>id</arg>.</par>
		"""
		s = str(id)
		if self.color:
			s = self.TextID(s)
		return s

	def __setitem__(self, id, target):
		"""
		<par>return the target with the id <arg>name</arg>.</par>
		"""
		if self.targets.has_key(id):
			self.reportWarning(RedefinedTargetWarning(id), 5)
		self.targets[id] = target

	def __getitem__(self, id):
		"""
		<par>return the target with the id <arg>id</arg>.</par>
		"""
		try:
			return self.targets[id]
		except KeyError:
			try:
				return self.targets[url.URL(id)] # retry with a filename
			except KeyError:
				return self.targets[url.here()/id] # retry with a absolute filename

	def __len__(self):
		"""
		<par>Return the number of targets.</par>
		"""
		return len(self.targets)

	def has(self, id):
		"""
		<par>Return whether the target with the id <arg>id</arg>
		exists in the project.</par>
		"""
		return self.targets.has_key(id)

	def build(self, *targets):
		"""
		<par>build all targets in <arg>targets</arg>.</par>
		
		<par>Items in <arg>targets</arg> must be <pyref class="Target"><class>Target</class></pyref>
		instances or target ids. If an id can't be found, it will be wrapped in a
		<pyref module="url" class="URL"><class>URL</class></pyref> instance and retried.</par>
		"""
		t1 = DateTime.now()
		self.rebuilt = 0
		self.failed = 0
		for target in targets:
			if not isinstance(target, Target):
				target = self[target]
			target.update()
		t2 = DateTime.now()
		if self.showSummary:
			self.report("build %s\n%s%s (%d targets; %d rebuilt; %d failed)\n" % (self.__class__.__name__, indent, self.secondsStr(t2-t1), len(self.targets), self.rebuilt, self.failed))
		self.rebuilt = 0
		self.failed = 0

	def buildWithArgs(self, args):
		(options, args) = getopt.getopt(args, "t:s:a:m:z:ic", ["showtargets=", "showsources=", "showactions=", "showmessages=", "showsummary=", "ignore", "color"])

		for (option, value) in options:
			if option=="-t" or option=="--showtargets":
				self.showTargets = int(value)
			if option=="-s" or option=="--showsources":
				self.showSources = int(value)
			if option=="-a" or option=="--showactions":
				self.showActions = int(value)
			if option=="-m" or option=="--showmessages":
				self.showMessages = int(value)
			if option=="-z" or option=="--showsummary":
				self.showSummary = int(value)
			if option=="-i" or option=="--ignore":
				self.ignoreExceptions = 1
			if option=="-c" or option=="--color":
				self.color = 1
		self.build(*args)

	def report(self, text):
		sys.stderr.write(text)

	def reportError(self, text):
		self.report(text)

	def reportWarning(self, warning, stacklevel):
		warnings.warn(str(warning), stacklevel=stacklevel) # FIXME: Drop the str() in Python 2.3

